// -*- mode: cpp; mode: fold -*-
// Description								/*{{{*/
// $Id: connect.cc,v 1.10.2.1 2004/01/16 18:58:50 mdz Exp $
/* ######################################################################

   Connect - Replacement connect call

   This was originally authored by Jason Gunthorpe <jgg@debian.org>
   and is placed in the Public Domain, do with it what you will.
      
   ##################################################################### */
									/*}}}*/
// Include Files							/*{{{*/
#include <config.h>

#include <apt-pkg/error.h>
#include <apt-pkg/fileutl.h>
#include <apt-pkg/strutl.h>
#include <apt-pkg/acquire-method.h>
#include <apt-pkg/configuration.h>

#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <sstream>

#include<set>
#include<string>

// Internet stuff
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "connect.h"
#include "rfc2553emu.h"
#include <apti18n.h>
									/*}}}*/

static std::string LastHost;
static int LastPort = 0;
static struct addrinfo *LastHostAddr = 0;
static struct addrinfo *LastUsed = 0;
static int nbHosts = 0;
static int use_parallel_connect = 0;

// Set of IP/hostnames that we timed out before or couldn't resolve
static std::set<std::string> bad_addr;

// RotateDNS - Select a new server from a DNS rotation			/*{{{*/
// ---------------------------------------------------------------------
/* This is called during certain errors in order to recover by selecting a 
   new server */
void RotateDNS()
{
   if (LastUsed != 0 && LastUsed->ai_next != 0)
      LastUsed = LastUsed->ai_next;
   else
      LastUsed = LastHostAddr;
}
									/*}}}*/
// DoConnect - Attempt a connect operation				/*{{{*/
// ---------------------------------------------------------------------
/* This helper function attempts a connection to a single address. */
static bool DoConnect(struct addrinfo *Addr,std::string Host,
		      unsigned long TimeOut,int &Fd,pkgAcqMethod *Owner)
{
   // Show a status indicator
   char Name[NI_MAXHOST];
   char Service[NI_MAXSERV];
   
   Name[0] = 0;   
   Service[0] = 0;
   getnameinfo(Addr->ai_addr,Addr->ai_addrlen,
	       Name,sizeof(Name),Service,sizeof(Service),
	       NI_NUMERICHOST|NI_NUMERICSERV);
   Owner->Status(_("Connecting to %s (%s)"),Host.c_str(),Name);

   // if that addr did timeout before, we do not try it again
   if(bad_addr.find(std::string(Name)) != bad_addr.end())
      return false;

   /* If this is an IP rotation store the IP we are using.. If something goes
      wrong this will get tacked onto the end of the error message */
   if (LastHostAddr->ai_next != 0)
   {
      std::stringstream ss;
      ioprintf(ss, _("[IP: %s %s]"),Name,Service);
      Owner->SetIP(ss.str());
   }
      
   // Get a socket
   if ((Fd = socket(Addr->ai_family,Addr->ai_socktype,
		    Addr->ai_protocol)) < 0)
      return _error->Errno("socket",_("Could not create a socket for %s (f=%u t=%u p=%u)"),
			   Name,Addr->ai_family,Addr->ai_socktype,Addr->ai_protocol);
   
   SetNonBlock(Fd,true);
   if (connect(Fd,Addr->ai_addr,Addr->ai_addrlen) < 0 &&
       errno != EINPROGRESS)
      return _error->Errno("connect",_("Cannot initiate the connection "
			   "to %s:%s (%s)."),Host.c_str(),Service,Name);
   
   /* This implements a timeout for connect by opening the connection
      nonblocking */
   if (WaitFd(Fd,true,TimeOut) == false) {
      bad_addr.insert(bad_addr.begin(), std::string(Name));
      Owner->SetFailReason("Timeout");
      return _error->Error(_("Could not connect to %s:%s (%s), "
			   "connection timed out"),Host.c_str(),Service,Name);
   }

   // Check the socket for an error condition
   unsigned int Err;
   unsigned int Len = sizeof(Err);
   if (getsockopt(Fd,SOL_SOCKET,SO_ERROR,&Err,&Len) != 0)
      return _error->Errno("getsockopt",_("Failed"));
   
   if (Err != 0)
   {
      errno = Err;
      if(errno == ECONNREFUSED)
         Owner->SetFailReason("ConnectionRefused");
      else if (errno == ETIMEDOUT)
	 Owner->SetFailReason("ConnectionTimedOut");
      bad_addr.insert(bad_addr.begin(), std::string(Name));
      return _error->Errno("connect",_("Could not connect to %s:%s (%s)."),Host.c_str(),
			   Service,Name);
   }
   
   return true;
}
									/*}}}*/
// connect_to_host - Connect to a server				/*{{{*/
// ---------------------------------------------------------------------
/* Performs a connection to the server */
#define TIMEOUT 500
static bool
connect_to_host(struct addrinfo *StartHost,struct addrinfo *FirstHostAddr,
		std::string Host,int count,int &Fd,
		unsigned long TimeOut,pkgAcqMethod * Owner) {
  struct addrinfo *Addr, *NextAddr, **Addrs = NULL;
  int fd = -1, n, i, j, max = -1, *fds;
  struct timeval *timeout, timeout0 = { 0, TIMEOUT * 1000},
    timeout1 = { 0, 0 };
  timeout1.tv_sec = TimeOut;
  fd_set fdset, wrset;
  // Show a status indicator
  char Name[NI_MAXHOST];
  char Service[NI_MAXSERV];
  
  fds = new int[count]();//calloc(count, sizeof(*fds));
  if (fds == NULL) {
    _error->Errno("calloc",_("Cannot allocate %d file descriptors"),count);
    return false;
  }
  Addrs = new struct addrinfo *[count]();
  if (Addrs == NULL) {
    _error->Errno("calloc",_("Cannot allocate %d pointers"),count);
    free(fds);
    return false;
  }
  FD_ZERO(&fdset);
  for(Addr = StartHost, i = 0, count = 0; Addr != NULL; Addr = NextAddr) {
    NextAddr = Addr;
  
    // Ignore UNIX domain sockets
    do {
      NextAddr = NextAddr->ai_next;
    } while (NextAddr != 0 && NextAddr->ai_family == AF_UNIX);

    /* If we reached the end of the search list then wrap around to the
       start */
    if (NextAddr == 0)
	NextAddr = FirstHostAddr;
      
    // Reached the end of the search cycle
    if (NextAddr == StartHost)
      NextAddr = NULL;


    Name[0] = 0;
    Service[0] = 0;
    getnameinfo(Addr->ai_addr,Addr->ai_addrlen,
		Name,sizeof(Name),Service,sizeof(Service),
		NI_NUMERICHOST|NI_NUMERICSERV);
    Owner->Status(_("Connecting to %s (%s)"),Host.c_str(),Name);

    fd = socket(Addr->ai_family, Addr->ai_socktype, Addr->ai_protocol);
    if (fd == -1) {
      /*
       * If AI_ADDRCONFIG is not supported we will get
       * EAFNOSUPPORT returned.  Behave as if the address
       * was not there.
       */
      if (errno != EAFNOSUPPORT) {
	_error->Errno("socket",_("Could not create a socket for %s (f=%u t=%u p=%u)"),
		      Name,Addr->ai_family,Addr->ai_socktype,Addr->ai_protocol);
      } else if (NextAddr != NULL)
	continue;
    } else if (fd >= FD_SETSIZE) {
      close(fd);
    } else {
      SetNonBlock(fd,true);
      if (connect(fd, Addr->ai_addr, Addr->ai_addrlen) == -1) {
	if (errno != EINPROGRESS) {
	  _error->Errno("connect",_("Cannot initiate the connection "
				    "to %s:%s (%s)."),Host.c_str(),Service,Name);
	  close(fd);
	} else {
	  /*
	   * Record the information for this descriptor.
	   */
	  fds[i] = fd;
	  Addrs[i] = Addr;
	  FD_SET(fd, &fdset);
	  if (max == -1 || fd > max)
	    max = fd;
	  count++;
	  i++;
	}
      } else  {
	/*
	 * We connected without blocking.
	 */
	LastUsed = Addr;
	goto done;
      }
    }

    if (count == 0)
      continue;
    
    //assert(max != -1);
    do {
      if (Addr->ai_next != NULL)
	timeout = &timeout0;
      else
	timeout = &timeout1;
      
      /* The write bit is set on both success and failure. */
      wrset = fdset;
      n = select(max + 1, NULL, &wrset, NULL, timeout);
      if (n == 0) {
	timeout0.tv_usec >>= 1;
	break;
      }
      if (n < 0) {
	if (errno == EAGAIN || errno == EINTR)
	  continue;
	_error->Errno("select", _("Cannot select"));
	fd = -1;
	goto done;
      }
      for (fd = 0; fd <= max; fd++) {
	if (FD_ISSET(fd, &wrset)) {
	  socklen_t len;
	  int err;
	  for (j = 0; j < i; j++)
	    if (fds[j] == fd)
	      break;
	  //assert(j < i);
	  /*
	   * Test to see if the connect
	   * succeeded.
	   */
	  len = sizeof(err);
	  n = getsockopt(fd, SOL_SOCKET,
			 SO_ERROR, &err, &len);
	  if (n != 0 || err != 0) {
	    if (n != 0) {
	      _error->Errno("getsockopt",_("Failed"));
	    } else {
	      errno = err;
	      if(errno == ECONNREFUSED)
		Owner->SetFailReason("ConnectionRefused");
	      else if (errno == ETIMEDOUT)
		Owner->SetFailReason("ConnectionTimedOut");
	      _error->Errno("connect",_("Could not connect to %s:%s (%s)."),
			    Host.c_str(),Service,Name);
	    }
	    close(fd);
	    FD_CLR(fd, &fdset);
	    fds[j] = -1;
	    count--;
	    continue;
	  }
	  /* Connect succeeded. */
	  LastUsed = Addrs[j];
	  goto done;
	}
      }
    } while (timeout == &timeout1 && count != 0);
  }
  
  /* We failed to connect. */
  fd = -1;
  
 done:
  /* Close all other descriptors we have created. */
  for (j = 0; j < i; j++)
    if (fds[j] != fd && fds[j] != -1) {
      close(fds[j]);
    }
  
  /* Free everything. */
  if (fds) delete[](fds);
  if (Addrs) delete[](Addrs);
  
  if (fd != -1) {
    Fd=fd;
    _error->Discard();
    return true;
  }

  return false;
}
									/*}}}*/
// Connect - Connect to a server					/*{{{*/
// ---------------------------------------------------------------------
/* Performs a connection to the server */
bool Connect(std::string Host,int Port,const char *Service,int DefPort,int &Fd,
	     unsigned long TimeOut,pkgAcqMethod *Owner)
{
   if (_error->PendingError() == true)
      return false;

   // Convert the port name/number
   char ServStr[300];
   if (Port != 0)
      snprintf(ServStr,sizeof(ServStr),"%u",Port);
   else
      snprintf(ServStr,sizeof(ServStr),"%s",Service);
   
   /* We used a cached address record.. Yes this is against the spec but
      the way we have setup our rotating dns suggests that this is more
      sensible */
   if (LastHost != Host || LastPort != Port)
   {
      Owner->Status(_("Connecting to %s"),Host.c_str());

      // Free the old address structure
      if (LastHostAddr != 0)
      {
	 freeaddrinfo(LastHostAddr);
	 LastHostAddr = 0;
	 LastUsed = 0;
	 nbHosts = 0;
      }
      
      // We only understand SOCK_STREAM sockets.
      struct addrinfo Hints;
      memset(&Hints,0,sizeof(Hints));
      Hints.ai_socktype = SOCK_STREAM;
      Hints.ai_flags = AI_ADDRCONFIG;
      Hints.ai_protocol = 0;
      
      if(_config->FindB("Acquire::ForceIPv4", false) == true)
         Hints.ai_family = AF_INET;
      else if(_config->FindB("Acquire::ForceIPv6", false) == true)
         Hints.ai_family = AF_INET6;
      else
         Hints.ai_family = AF_UNSPEC;

      // if we couldn't resolve the host before, we don't try now
      if(bad_addr.find(Host) != bad_addr.end()) 
	 return _error->Error(_("Could not resolve '%s'"),Host.c_str());

      // Resolve both the host and service simultaneously
      while (1)
      {
	 int Res;
	 if ((Res = getaddrinfo(Host.c_str(),ServStr,&Hints,&LastHostAddr)) != 0 ||
	     LastHostAddr == 0)
	 {
	    if (Res == EAI_NONAME || Res == EAI_SERVICE)
	    {
	       if (DefPort != 0)
	       {
		  snprintf(ServStr,sizeof(ServStr),"%u",DefPort);
		  DefPort = 0;
		  continue;
	       }
	       bad_addr.insert(bad_addr.begin(), Host);
	       Owner->SetFailReason("ResolveFailure");
	       return _error->Error(_("Could not resolve '%s'"),Host.c_str());
	    }
	    
	    if (Res == EAI_AGAIN)
	    {
	       Owner->SetFailReason("TmpResolveFailure");
	       return _error->Error(_("Temporary failure resolving '%s'"),
				    Host.c_str());
	    }
	    if (Res == EAI_SYSTEM)
	       return _error->Errno("getaddrinfo", _("System error resolving '%s:%s'"),
                                      Host.c_str(),ServStr);
	    return _error->Error(_("Something wicked happened resolving '%s:%s' (%i - %s)"),
				 Host.c_str(),ServStr,Res,gai_strerror(Res));
	 }
	 break;
      }
      
      LastHost = Host;
      LastPort = Port;

      /*
       * Work out how many possible IP we could use.
       */
      struct addrinfo *res;
      for (res = LastHostAddr; res; res = res->ai_next) {
	// Ignore UNIX domain sockets
	if (res->ai_family == AF_UNIX)
	  continue;
	nbHosts++;
      }
   }

   // When we have an IP rotation stay with the last IP.
   struct addrinfo *CurHost = LastHostAddr;
   if (LastUsed != 0)
       CurHost = LastUsed;
   
   if (use_parallel_connect == 0) {
     use_parallel_connect=1;
     if (getenv("APT_SEQUENTIAL_CONNECT") != NULL)
       use_parallel_connect=-1;
   }
   if (use_parallel_connect == 1) {
     return connect_to_host(CurHost,LastHostAddr,Host,nbHosts,Fd,TimeOut,Owner);
   }

   while (CurHost != 0)
   {
      if (DoConnect(CurHost,Host,TimeOut,Fd,Owner) == true)
      {
	 LastUsed = CurHost;
	 return true;
      }      
      close(Fd);
      Fd = -1;
      
      // Ignore UNIX domain sockets
      do
      {
	 CurHost = CurHost->ai_next;
      }
      while (CurHost != 0 && CurHost->ai_family == AF_UNIX);

      /* If we reached the end of the search list then wrap around to the
         start */
      if (CurHost == 0 && LastUsed != 0)
	 CurHost = LastHostAddr;
      
      // Reached the end of the search cycle
      if (CurHost == LastUsed)
	 break;
      
      if (CurHost != 0)
	 _error->Discard();
   }   

   if (_error->PendingError() == true)
      return false;   
   return _error->Error(_("Unable to connect to %s:%s:"),Host.c_str(),ServStr);
}
									/*}}}*/
